From Hammer Require Import Hammer.
Require Import Coq.Program.Basics.
Require Import Coq.Program.Tactics.
Require Import Unicode.Utf8.
Require Import Coq.Logic.Classical_Prop.
Require Import Coq.Logic.PropExtensionality.
Require Import Coq.Relations.Relation_Definitions.
Require Import Coq.Logic.ClassicalEpsilon.
Require Import Coq.Logic.Classical_Pred_Type.
Require Import Setoid Morphisms.
Require Import Coq.Classes.Morphisms.
Require Import Coq.Classes.RelationClasses.
Require Import Coq.Strings.String.
Require Import Coq.Logic.FinFun.
Require Import Mereology_on_Onto_sets_v10.

From Hammer Require Reconstr.
Set Hammer ReconstrLimit 25.

Hammer_version.
Hammer_objects.
Set Hammer ATPLimit 180.

Module Type Mereo_signature (Mereo : DST_signature).
Import Mereo.

Parameter isBall : object -> Prop.
Definition balls :=  Caract (fun A:object => isBall A).

(*========================================================*)

(** The space of mereogeometry is built on balls  **)

Definition Gspace : N :=  Caract (fun A:object => IF_then_else (η (ι A) (klass balls)) True False).

Lemma GeomSpace : forall A, η A Gspace <-> η A (klass balls).
Proof.
intro A;split.
- intro;unfold η in H;destruct H as [[x H] H20];assert (H21:set_eq (ι x) A /\ incl A Gspace).
  -- split;assumption.
  -- apply incl_in_singleton in H21;unfold In in H21;unfold Gspace in H21;unfold IF_then_else in H21;simpl in H21;assert (H1:True).
     --- auto.
     --- rewrite <-H21 in H1;clear H21;destruct H1.
         ---- destruct H0;apply rewl_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ].
         ---- destruct H0;contradiction.
- intro;assert (H0:=H);destruct H as [[x H] H1];unfold η;split.
  -- destruct H0;assumption.
  -- apply in_in_singleton with (A:=x);split.
     --- assumption.
     --- unfold In;unfold Gspace;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;split;[ hauto use: in_singleton, equiv_singleton, N9 unfold: set_eq, incl, η | auto ].
Qed.

Axiom GS : exists E, η E Gspace.

Lemma in_GeomSpace : forall A, η A balls -> η A (el Gspace).
Proof.
intros A H;assert (set_eq Gspace (klass balls)).
- hfcrush use: GeomSpace, N22.
- apply MieT22 in H;apply MieT18 with (B:=klass balls);sfirstorder.
Qed.

Lemma ball_exists : !balls.
Proof.
unfold exists_at_least;assert (exists E, η E Gspace);[ apply GS | 
destruct H as [E H];rewrite GeomSpace in H;apply MieT2 in H;destruct H as [A H];exists A;assumption ].
Qed.

(** there exists at least an object that is a balls **)

Lemma Ball_existence : !balls <-> exists E, η E Gspace.
Proof.
hecrush use: MieT25, GeomSpace unfold: exists_at_least.
Qed.

Lemma single_Geospace : exists_at_most Gspace.
Proof.
unfold exists_at_most;intros A B [H1 H2];rewrite GeomSpace in H1, H2;apply klUniq with (a:=balls);split;assumption.
Qed.

Lemma space_is_klass : forall A, η A Gspace -> set_eq A Gspace.
Proof.
intro A;intro H;assert (Individual A /\ Individual Gspace).
- hauto lq: on use: single_Geospace, N2, N1, OntoT265.
- apply singular_eq_dec in H0;rewrite H0;strivial use: single_Geospace, OntoT282 unfold: singular_eq.
Qed.

Lemma indiv_Gspace : forall A, η A Gspace -> Individual Gspace.
Proof.
hauto lq: on use: space_is_klass, N2, N1, indiv_singleton_propag.
Qed.

(** Definition of a solid as a ball's collection **)


Definition Solid : N :=  Caract (fun A:object => IF_then_else (exists a, a ⊆ balls /\ η (ι A) (klass a)) True False).

Lemma is_solid : forall A, η A Solid <-> (Individual A /\ exists b, b ⊆ balls /\ η A (klass b)).
Proof.
intro A;split.
- intro H;assert (H10:=H);unfold η in H;destruct H as [[x H] H20];assert (H21:set_eq (ι x) A /\ incl A Solid);[ split;assumption |
  apply incl_in_singleton in H21;unfold In in H21;unfold Solid in H21;unfold IF_then_else in H21;simpl in H21;assert (H1:True);[ auto |
  split;[ destruct H10;assumption | rewrite <-H21 in H1;clear H21;destruct H1;[ destruct H0 as [[a [H2 H3]] H0];exists a;
  hauto lq: on use: rewl_singleton_in_η, singular_eq_dec, singular_eq_eq_obj unfold: η, klass, ι, singular_eq | destruct H0;contradiction ]]]].
- intro;assert (H0:=H);destruct H as [[x H] H1];unfold η;split;[ destruct H0;assumption | apply in_in_singleton with (A:=x);split;[ assumption |
 unfold In;unfold Solid;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto | intro;classical_left;split;[ 
 destruct H0 as [H0 [a [H4 H5]]];exists a;split;[ assumption | apply rewr_singleton_in_η with (σ:=A);split;[ | apply set_eq_sym ];assumption] | auto ]]]].
Qed.


Lemma solid_as_coll : forall A, η A Solid <-> η A (coll balls).
Proof.
qauto depth: 4 l: on use: L19a, M18, is_solid, Klass.
Qed.

(** any ball is a solid but the converse is not true **)

Lemma any_ball_is_a_solid : forall A, η A balls -> η A Solid.
Proof.
hcrush use: solid_as_coll, SinXIII unfold: weakInclusion.
Qed. 

(** The geometric space is also a solid **)

Lemma whole_space_is_a_solid : forall A, η A Gspace -> η A Solid.
Proof.
hauto use: GeomSpace, OntoT49, N1, OntoT14, is_solid.
Qed.

Lemma Solid_el_of_Gspace : forall A, η A Solid -> η A (el Gspace).
Proof.
intros A H;rewrite solid_as_coll in H;apply SinXXVI in H;assert (set_eq Gspace (klass balls));[
hfcrush use: GeomSpace, N22 | apply MieT18 with  (B:= klass balls);sauto ].
Qed.

Lemma any_solid_includes_balls : forall A, η A Solid -> exists C, η C balls /\ η C (el A).
Proof.
hfcrush use: SinXXI, solid_as_coll.
Qed.

Lemma any_balls_included_in_solid : forall A, η A balls -> exists C, η C Solid /\ η C (el A).
Proof.
hauto lq: on use: MieT7, any_ball_is_a_solid, N2.
Qed.

(** A is externally tangent to B **)

Definition ET (B:N) : N :=
    Caract (fun A:object => IF_then_else (In balls A /\ η B balls /\ η (ι A)(ext B) /\ (forall x y, η x balls /\ η y balls /\ 
                                          η (ι A) (el x) /\ η x (ext B) /\ η (ι A)(el y) /\ η y (ext B) -> 
                                         (η x (el y) \/ η y (el x)))) True False).

Lemma extTanTo : forall A B, η A (ET B) <-> (η A balls /\ η B balls /\ η A (ext B) /\ (forall x y, η x balls /\ η y balls /\ 
                                          η A (el x) /\ η x (ext B) /\ η A (el y) /\ η y (ext B) -> (η x (el y) \/ η y (el x)))).
Proof.
intros A B;split.
- intro;solve_op_in_hyp H ET x;destruct H2.
  -- destruct H1 as [[H3 [H4 [H5 H6]]] H2];split.
     --- hauto lq: on use: in_in_singleton unfold: η.
     --- split.
         ---- assumption.
         ---- split.
              ----- hauto lq: on use: in_in_singleton, N11 unfold: η.
              ----- intros P Q H7;specialize (H6 P Q);subst_set_eq_in_hyp H el P;rewrite H1 in H6.
                    subst_set_eq_in_hyp H el Q;rewrite H8 in H6;apply H6 in H7;assumption.
  -- destruct H1;contradiction.
- intro;destruct H as [H3 [H4 [H5 H6]]];unfold η;split.
  -- destruct H3;assumption.
  -- destruct H3 as [[x H1] H2];apply in_in_singleton with (A:=x);split.
     --- assumption.
     --- solve_op_in_goal ET;split.
         ---- split.
              ----- qauto use: in_singleton unfold: set_eq, incl.
              ----- split.
                    + assumption.
                    + split.
                      ++ sfirstorder use: N9, In_singleton_incl_equiv, η_singleton_l, equiv_singleton unfold: η, incl.
                      ++ intros P Q H7;specialize (H6 P Q);subst_set_eq_in_hyp H1 el P;rewrite H0 in H7;subst_set_eq_in_hyp H1 el Q.
                         rewrite H3 in H7;apply H6 in H7;assumption.
         ---- auto.
Qed.


Lemma ET_singleton : forall (Q:N)(A :object), set_eq Q (ι A) -> (set_eq (ET (ι A))(ET Q)).
Proof.
intros Q x H1;unfold set_eq;intro y;apply propositional_extensionality;split. 
- intro H;unfold In in H;unfold ET in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ET;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;destruct H as [[H2 [H3 [H4 H5]]] H6];split.
              ----- split;[ assumption |
                    split;[ apply rewl_singleton_in_η with (A:=x);split;assumption |
                    split;[ apply rewl_ext_singleton_in_η  with (A:=x);split;assumption |
                    intros u v [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                    split;[ assumption | split;[ assumption | split;[ apply rewr_ext_singleton_in_η  with (ϕ:=Q);split;assumption |
                    split;[ assumption | apply rewr_ext_singleton_in_η with (ϕ:=Q);split;assumption ]]]]]]]].
              ----- auto. 
     --- destruct H;contradiction.
- intro H;unfold In in H;unfold ET in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ET;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;destruct H as [[H2 [H3 [H4 H5]]] H6];split.
              ----- split;[ assumption |
                    split;[ apply rewr_singleton_in_η with (σ:=Q);split;assumption |
                    split;[ apply rewr_ext_singleton_in_η  with (ϕ:=Q);split;assumption |
                    intros u v [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                    split;[ assumption |
                    split;[ assumption |
                    split;[ apply rewl_ext_singleton_in_η with (A:=x);split;assumption |
                    split;[ assumption | apply rewl_ext_singleton_in_η with (A:=x);split;assumption ]]]]]]]].
              ----- auto.
     --- destruct H;contradiction.
Qed.

Lemma et_subst : forall (A B :N), A ≡ B -> (set_eq (ET A)(ET B)).
Proof.
intros A B H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v));[ split;apply N9 | apply singular_eq_dec in H;assert (set_eq (ι u) (ι v)); sfirstorder ]. 
- rewrite <-H in K10;apply set_eq_sym in K10;apply ET_singleton in K10;apply set_eq_sym in K9;apply ET_singleton in K9;assert (set_eq (ET A) (ET B));[
  sfirstorder | assumption ].
Qed.


Lemma ET_is_ext : forall A B, η A (ET B) -> η A (ext B).
Proof.
strivial use: extTanTo.
Qed.

Lemma ET_to_el : forall A B, η A (ET B) -> exists A', η A (el A') /\ η A' (ET B).
Proof.
strivial use: SinII unfold: η.
Qed.


Lemma not_ET_to_ov : forall A B E, η E balls /\ η A (pt E) /\ η A (ET B) /\ ~η E (ET B) -> ~η E (ext B).
Proof.
intros A B E [H1 [H2 [H3 H4]]];assert (K1:=H3);rewrite extTanTo in H4;apply not_and_or in H4;destruct H4.
- contradiction.
- rewrite extTanTo in H3;destruct H3 as [H3 [K2 [K3 K4]]];apply not_and_or in H;destruct H.
  -- contradiction.
  -- apply not_and_or in H;destruct H.
     --- assumption.
     --- intro;apply H;intros X Y [K5 [K6 [K7 [K8 [K9 K0]]]]];apply K4;split;[ assumption |
         split;[ assumption |
         split;[ apply SinXVIII in H2;apply SinXXIII with (Q:=E);split;assumption |
         split;[ assumption | split;[ apply SinXVIII in H2;apply SinXXIII with (Q:=E);split;assumption | assumption ]]]]].
Qed.


(** A is internally tangent to B **)

Definition IT (B:N) : N :=
    Caract (fun A:object => IF_then_else (In balls A /\ η B balls /\ η (ι A)(pt B) /\ (forall x y, η x balls /\ η y balls /\ η (ι A)(el x) /\
                                η x (el B) /\ η (ι A)(el y) /\ η y (el B) -> (η x (el y) \/ η y (el x)))) True False).

Lemma intTanTo : forall A B, η A (IT B) <-> (η A balls /\ η B balls /\ η A (pt B) /\ (forall x y, η x balls /\ η y balls /\ η A (el x) /\
                                η x (el B) /\ η A (el y) /\ η y (el B) -> (η x (el y) \/ η y (el x)))).
Proof.
intros A B;split.
- intro;assert (H0:=H);solve_op_in_hyp H IT x;destruct H3.
  -- destruct H2 as [[H3 [H4 [H5 H6]]] H2];split.
     --- hauto lq: on use: in_in_singleton unfold: η.
     --- split.
         ---- assumption.
         ---- split.
              + hauto lq: on use: in_in_singleton, N11 unfold: η.
              + intros P Q H7;specialize (H6 P Q);apply H6;destruct H7 as [H7 [K1 [K2 [K3 [K4 K5]]]]];hfcrush use: equiv_singleton, indiv_singl_l unfold: incl, η.
  -- destruct H2;contradiction.
- intro;destruct H as [H3 [H4 [H5 H6]]];unfold η;split.
  -- destruct H3;assumption.
  -- destruct H3 as [[x H1] H2];apply in_in_singleton with (A:=x);split.
     --- assumption.
     --- solve_op_in_goal IT;split.
         ---- split;[ apply incl_in_singleton with (Φ:=A);split;assumption |
              split;[ assumption |
              split;[ apply rewr_singleton_in_η with (σ:=A);solve_functor |
              intros P Q H7;specialize (H6 P Q);subst_set_eq_in_hyp H1 el P;rewrite H0 in H7;
              subst_set_eq_in_hyp H1 el Q;rewrite H3 in H7;apply H6 in H7;assumption ]]].
         ---- auto.
Qed.

Lemma IT_singleton : forall (Q:N)(A :object), set_eq Q (ι A) -> (set_eq (IT (ι A))(IT Q)).
Proof.
intros Q x H1;unfold set_eq;intro y;apply propositional_extensionality;split. 
- intro H;unfold In in H;unfold IT in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold IT;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;destruct H as [[H2 [H3 [H4 H5]]] H6];split.
              ----- split.
                + assumption.
                + split;[ apply rewl_singleton_in_η with (A:=x);split;assumption |
                  split;[  apply MieT17 with (B:=ι x);split;[ assumption | apply set_eq_sym;assumption ] |
                  intros u v [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                  split;[ assumption | split;[ assumption | split;[ apply MieT18 with (B:=Q);split;assumption |
                  split;[ assumption | apply MieT18 with (B:=Q);split;assumption ]]]]]]].
              ----- auto. 
     --- destruct H;contradiction. 
- intro H;unfold In in H;unfold IT in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold IT;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;destruct H as [[H2 [H3 [H4 H5]]] H6];split.
              ----- split;[ assumption |
                    split;[ apply rewr_singleton_in_η with (σ:=Q);split;assumption |
                    split;[ apply MieT17 with (B:=Q);split;assumption |
                    intros u v [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption | split;[ assumption |
                    split;[ assumption | split;[ apply MieT18 with (B:=ι x);split;[ assumption | apply set_eq_sym;assumption ] |
                    split;[ assumption | apply MieT18 with (B:=ι x);split;[ assumption | apply set_eq_sym;assumption ]]]]]]]]].
              ----- auto.
     --- destruct H;contradiction.
Qed.

Lemma it_subst : forall (A B :N), A ≡ B -> (set_eq (IT A)(IT B)).
Proof.
intros A B H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v));[ split;apply N9 | apply singular_eq_dec in H;assert (set_eq (ι u) (ι v));sfirstorder ].
- apply set_eq_sym in K9;apply IT_singleton in K9;apply set_eq_sym in K10;apply IT_singleton in K10;apply set_eq_trans with (y:=IT (ι v));[
  apply set_eq_trans with (y:=IT (ι u));[ apply set_eq_sym;assumption | sfirstorder ] | assumption ].
Qed.

Lemma IT_is_pt : forall A B, η A (IT B) -> η A (pt B).
Proof.
strivial use: intTanTo.
Qed.

Lemma et_bi_to_it : forall A B C,  η A (ET B) /\ η C (pt A) /\ η C (ET B) -> η C (IT A).
Proof.
intros A B C [H1 [H2 H3]];rewrite extTanTo in H3, H1;destruct H1 as [G1 [G2[ G3 G4]]].
destruct H3 as [K1 [K2[ K3 K4]]];rewrite intTanTo;split.
- sfirstorder. 
- split;[ assumption | split;[ assumption | intros P Q [H4 [H5 [H6 [H7 [H8 H9]]]]];apply K4;specialize (G4 P Q); split;[ assumption |
 split;[ assumption | split;[ assumption | split;[ apply SinXXXII;apply SinCXXII with (Q:=A);split;[apply SinXXXII |];assumption |
 split;[ assumption | apply SinXXXII;apply SinCXXII with (Q:=A);split;[apply SinXXXII |];assumption ]]]]]]].
Qed.

Lemma it_bi_to_et : forall A B C, η B (IT A) /\ η C (ET B) /\ η C (ext A) -> η C (ET A).
Proof.
intros A X V H;destruct H as [h1 [h2 h3]];rewrite intTanTo in h1;rewrite extTanTo in *;destruct h1 as [H1 [H2 [H3 H4]]].
destruct h2 as [K1 [K2 [K3 K4]]];split;[ assumption |
split;[ assumption | split;[ assumption | intros x y [H5 [H6 [H7 [H8 [H9 H10]]]]];apply K4;split;[ assumption |
split;[ assumption | split;[ assumption | split;[  apply SinCXXII with (Q:=A);split;[ assumption | strivial use: part ] |
split;[ assumption | apply SinCXXII with (Q:=A);split;[ assumption | strivial use: part ]]]]]]]]].
Qed.

Lemma et_dist : forall A F X, η A balls /\ η F (ET X) /\ η F (pt A) /\ η A (ext X) -> η A (ET X).
Proof.
intros. destruct H as [H1 [H2 [H3 H4]]]. rewrite extTanTo;split.
- assumption.
- split.
  -- rewrite extTanTo in H2;destruct H2 as [K1 [K2 K3]];assumption.
  -- split.
     --- assumption.
     --- intros;destruct H as [K1 [K2 [K3 [K4 [K5 K6]]]]];rewrite extTanTo in H2;destruct H2 as [L1 [L2 [L3 L4]]].
         apply L4;split;[ assumption |
         split;[ assumption |
         split;[ apply SinXXIII with (Q:=A);split;[ apply SinXVIII in H3;assumption | assumption ] |
         split;[ assumption |
         split;[ apply SinXXIII with (Q:=A);split;[ apply SinXVIII in H3;assumption | assumption ] | assumption ]]]]].
Qed.

Lemma dist_ET : forall A B C D, η B (IT A) /\ η C (ET B) /\ η D balls /\ η C (el D) /\ η D (ext A) -> η D (ET B).
Proof.
intros A B C D [H1 [H2 [H3 [H4 H5]]]];rewrite extTanTo in *;destruct H2 as [K1 [K2 [K3 K4]]];split;[ assumption |
split;[ assumption | split;[ rewrite intTanTo in H1;destruct H1 as [H1 [H6 [H7 H8]]];apply SinCXXII with (Q:=A);split;[ 
assumption | strivial use: part ] | intros X Y [H6 [H7 H8]];apply K4;clear K4;split;[ assumption |
split;[ assumption | destruct H8 as [J1 [J2 [J3 J4]]];split;[ apply SinXXIII with (Q:=D);split;assumption |
split;[ assumption | split;[ apply SinXXIII with (Q:=D);split;assumption | assumption ]]]]]]]].
Qed.

(** A is externally diametrically tangent to B relative to C **)


Definition ED (C B:N) : N := Caract (fun A:object => IF_then_else (η (ι A)(ET C) /\ η B (ET C) /\ (forall P Q, η P balls /\ η Q balls /\
                                                   η (ι A)(el P) /\ η P (ext C) /\ η B (el Q) /\ η Q (ext C) -> η P (ext Q))) True False).


Lemma diamExtTanTo : forall A B C, η A (ED C B) <-> (η A (ET C) /\ η B (ET C) /\ (forall P Q, η P balls /\ η Q balls /\
                                                   η A (el P) /\ η P (ext C) /\ η B (el Q) /\ η Q (ext C) -> η P (ext Q))).
Proof.
intros A B C;split.
- intro;assert (H0:=H);unfold η in H;destruct H as [[x H] H1];assert (H2:In (ED C B) x).
  -- apply incl_in_singleton with (Φ:=A);split;assumption.
  -- unfold In in H2;unfold ED in H2;unfold IF_then_else in H2;simpl in H2;assert (H3:True).
     --- auto.
     --- rewrite <-H2 in H3;clear H2;destruct H3.
         ---- destruct H2 as [[H3 [H4 H5]] H2];split;[ apply rewl_singleton_in_η with (A:=x);solve_functor |
              split;[ assumption |
              intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;clear H5;split;[ assumption |
              split;[ assumption | split;[ apply rewr_singleton_in_η with (σ:=A);solve_functor |
              split;[ assumption | split;assumption ]]]]]].
         ---- destruct H2;contradiction.
- intro; destruct H as [H3 [H4 H5]];assert (H0:=H3);destruct H0 as [H0 H7];unfold η;split.
  -- assumption.
  -- destruct H0 as [x H0];apply in_in_singleton with (A:=x);split.
     --- assumption.
     --- unfold In;unfold ED;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;split.
              ----- split;[ apply set_eq_sym in H0;apply rewr_singleton_in_η with (σ:=A);split;assumption |
                    split;[ assumption | intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption | split;[ assumption |
                    split;[ apply rewl_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ] |
                    split;[ assumption |
                    split;assumption ]]]]]].
              ----- auto.
Qed.

Lemma edt_are_ext : forall A B C, η A (ED C B) -> η A (ext C) /\ η B (ext C).
Proof.
hauto lq: on use: diamExtTanTo, ET_is_ext.
Qed.

Lemma edt_is_ET : forall A B C, η A (ED C B) -> (η A (ET C) /\ η B (ET C)).
Proof.
strivial use: diamExtTanTo.
Qed.

Lemma ED_singleton_l : forall (B C:N)(x :object), set_eq B (ι x) -> (set_eq (ED C (ι x))(ED C B)).
Proof. 
intros B C x H1;unfold set_eq;intro y;apply propositional_extensionality;split.
- intro H;unfold In in H;unfold ED in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ED;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;destruct H as [[H3 [H4 H5]] H2];split.
              ----- split;[ assumption |
                    split;[ apply rewl_singleton_in_η with (A:=x);split;assumption |
                    intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                    split;[ assumption |
                    split;[ assumption |
                    split;[ assumption |
                    split;[ apply rewr_singleton_in_η with (σ:=B);split;assumption | assumption ]]]]]]].
              ----- auto.
     --- destruct H;contradiction.
- intro H;unfold In in H;unfold ED in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ED;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;left;destruct H as [[H3 [H4 H5]] H2];split.
              -----  split;[ assumption | split;[ apply rewr_singleton_in_η with (σ:=B);split;assumption |
                     intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                     split;[ assumption |
                     split;[ assumption |
                     split;[ assumption |
                     split;[ strivial use: η_singleton unfold: η_inv | assumption ]]]]]]].
              ----- auto.
     --- destruct H;contradiction.
Qed.

Lemma ED_singleton_r : forall (B C:N)(x :object), set_eq B (ι x) -> (set_eq (ED (ι x) C)(ED B C)).
Proof.
intros B C x H1;unfold set_eq;intro y;apply propositional_extensionality;split.
- intro H;unfold In in H;unfold ED in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ED;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H3 [H4 H5]] H6];split. 
              ----- split;[ hcrush use: ET_singleton, ext_singl |
                  split;[ apply ET_singleton in H1;rewrite N22 in H1;rewrite <-H1;assumption |
                  intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                  split;[ assumption |
                  split;[ assumption | split;[ apply rewr_ext_singleton_in_η with (ϕ:=B);split;assumption |
                  split;[ assumption | apply rewr_ext_singleton_in_η with (ϕ:=B);split;assumption ]]]]]]].
              ----- auto .
      --- destruct H;contradiction.
- intro H;unfold In in H;unfold ED in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ED;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H3 [H4 H5]] H6];split.
              ----- split;[ apply ET_singleton in H1;rewrite N22 in H1;rewrite H1;assumption |
                    split;[ apply ET_singleton in H1;rewrite N22 in H1;rewrite H1;assumption |
                    intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];apply H5;split;[ assumption |
                    split;[ assumption |
                    split;[ assumption | split;[ apply rewl_ext_singleton_in_η with (A:=x);split;assumption |
                    split;[ assumption | apply rewl_ext_singleton_in_η with (A:=x);split;assumption ]]]]]]]. 
              ----- auto.
      --- destruct H;contradiction.
Qed.


Lemma ED_subst_r : forall (A B C:N), A ≡ B -> (set_eq (ED C A)(ED C B)).
Proof.
intros A B C H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v)).
  -- split;apply N9.
  -- apply singular_eq_dec in H;assert (set_eq (ι u) (ι v)).
     --- apply set_eq_trans with (y:=A);[ assumption | apply set_eq_trans with (y:=B);[ assumption | apply set_eq_sym;assumption ]].
     --- rewrite H in H0;assumption.
- rewrite <-H in K10;apply set_eq_sym in K10;apply ED_singleton_l with (C:=C) in K10;apply set_eq_sym in K9;apply ED_singleton_l with (C:=C) in K9.
  apply set_eq_trans with (y:=ED C (ι u));[ apply set_eq_sym;assumption | assumption ].
Qed.

Lemma ED_subst_l : forall (A B C:N), A ≡ B -> (set_eq (ED A C)(ED B C)).
Proof.
intros A B C H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v)).
  -- split;apply N9.
  -- apply singular_eq_dec in H;assert (set_eq (ι u) (ι v)).
     --- apply set_eq_trans with (y:=A);[ assumption | apply set_eq_trans with (y:=B);[ assumption | apply set_eq_sym;assumption ]].
     --- rewrite H in H0;assumption.
- rewrite <-H in K10;apply set_eq_sym in K10;apply ED_singleton_r with (C:=C) in K10;apply set_eq_sym in K9;apply ED_singleton_r with (C:=C) in K9.
  apply set_eq_trans with (y:=ED (ι u) C);[ apply set_eq_sym;assumption | assumption ].
Qed.


Lemma ED_sym : forall A B C, η A (ED C B) <-> η B (ED C A).
Proof.
intros A B C;split;( intro H;rewrite diamExtTanTo in *;destruct H as [H1 [H2 H3]];split;[ assumption |
                     split;[ sauto | intros P Q [K1 [K2 [K3 [K4 [K5 K6]]]]];hfcrush use: SinXXXII ]]).
Qed.


(** A is internally diametrically tangent to B relative to C **)

Definition ID (C B:N) : N := Caract (fun A:object => IF_then_else (η (ι A)(el C) /\ η B (el C) /\ (exists P Q R, η P balls /\ η Q balls /\
                                                   η R balls /\ η P (ED (ι A) Q) /\ η P (ED B R) /\ η Q (ED C R)) ) True False).


Lemma diamIntTan : forall A B C, η A (ID C B) <-> (η A (el C) /\ η B (el C) /\ (exists P Q R, η P balls /\ η Q balls /\
                                                   η R balls /\ η P (ED A Q) /\ η P (ED B R) /\ η Q (ED C R))).
Proof.
intros A B C;split.
- intro;unfold η in H;destruct H as [[x H] H1];assert (H2:In (ID C B) x).
  -- apply incl_in_singleton with (Φ:=A);split;assumption.
  -- unfold In in H2;unfold ID in H2;unfold IF_then_else in H2;simpl in H2;assert (H3:True).
     --- auto.
     --- rewrite <-H2 in H3;clear H2;destruct H3.
         ---- destruct H0 as [[H3 [H4 H5]] H0];split;[ strivial use: set_eq_sym, η_singleton unfold: η_inv |
              split;[ assumption |
              destruct H5 as [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]];exists E, F, G;split;[ assumption |
              split;[ assumption |
              split;[ assumption |
              split;[ apply set_eq_sym in H;apply ED_singleton_r with (C:=F) in H;rewrite N22 in H;rewrite <-H;assumption |
              split;assumption ]]]]]].
         ---- destruct H0;contradiction.
- intros [H1 [H2 H3]];unfold η;assert (H0:=H1);destruct H0 as [H0 H4];split.
  -- assumption.
  -- destruct H0 as [x H0];clear H4;apply in_in_singleton with (A:=x);split.
     --- assumption.
     --- unfold In;unfold ID;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;split;[ split;[
              apply rewr_singleton_in_η with (σ:=A);split;[ assumption | apply set_eq_sym;assumption ] |
              split;[ assumption |
              destruct H3 as [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]];exists E, F, G;split;[ assumption |
              split;[ assumption |
              split;[ assumption |
              split;[ apply set_eq_sym in H0;apply ED_singleton_r with (C:=F) in H0;rewrite N22 in H0;rewrite H0;assumption |
              split;assumption ]]]]]] |
              auto ].
Qed.

Lemma ID_singleton_l : forall (B C:N)(x :object), set_eq B (ι x) -> (set_eq (ID (ι x) C)(ID B C)).
Proof. 
intros B C x H1;unfold set_eq;intro y;apply propositional_extensionality;split.
- intro H;unfold In in H;unfold ID in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ID;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H3 [H4 H5]] H6];clear H2;split;[
              split;[ apply rewl_el_singleton_in_η with (A:=x);split;assumption |
              split;[ apply rewl_el_singleton_in_η with (A:=x);split;assumption |
              destruct H5 as [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]];exists E, F, G;split;[ assumption |
              split;[ assumption |
              split;[ assumption |
              split;[ assumption | split;[ assumption |
              apply ED_singleton_r with (C:=G) in H1;rewrite N22 in H1;rewrite <-H1;assumption ]]]]]]] | auto ].
     --- destruct H;contradiction.
- intro H;unfold In in H;unfold ID in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ID;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H3 [H4 H5]] H6]; clear H2;split;[
              split;[ apply rewr_el_singleton_in_η with (ϕ:=B);split;assumption |
              split;[ apply rewr_el_singleton_in_η with (ϕ:=B);split;assumption |
              destruct H5 as [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]];exists E, F, G;split;[ assumption |
              split;[ assumption |
              split;[ assumption |
              split;[ assumption |
              split;[ assumption | hcrush use: N22, set_eq_sym, ED_singleton_r ]]]]]]] | auto ].
     --- destruct H;contradiction.
Qed.


Lemma ID_singleton_r : forall (B C:N)(x :object), set_eq B (ι x) -> (set_eq (ID C (ι x))(ID C B)).
Proof. 
intros B C x H1;unfold set_eq;intro y;apply propositional_extensionality;split.
- intro H;unfold In in H;unfold ID in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ID;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H3 [H4 H5]] H6];clear H2;split;[ split;[
              assumption | split;[ apply rewl_singleton_in_η with (A:=x);split;assumption |
              destruct H5 as [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]];exists E, F, G;split;[
              assumption | split;[ assumption | split;[ assumption | split;[ assumption | split;[
              apply ED_singleton_r with (C:=G) in H1;rewrite N22 in H1;rewrite <-H1;assumption |
              assumption ]]]]]]] | auto ].
     --- destruct H;contradiction.
- intro H;unfold In in H;unfold ID in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold ID;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H3 [H4 H5]] H6];clear H2;split;[ split;[ assumption |
                split;[ apply rewr_singleton_in_η with (σ:=B);split;assumption |
                destruct H5 as [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]];exists E, F, G;split;[
                assumption | split;[ assumption | split;[ assumption | split;[ assumption |
                split;[ hcrush use: N22, set_eq_sym, ED_singleton_r | assumption ]]]]]]] | auto ].
     --- destruct H;contradiction.
Qed.

Lemma ID_subst_l : forall (A B C:N), A ≡ B -> (set_eq (ID A C)(ID B C)). 
Proof.
intros A B C H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v)).
  -- split;apply N9.
  -- apply singular_eq_dec in H;assert (set_eq (ι u) (ι v)).
     --- apply set_eq_trans with (y:=A);[ assumption | apply set_eq_trans with (y:=B);[ assumption | apply set_eq_sym;assumption ]].
     --- rewrite H in H0;assumption.
- rewrite <-H in K10;apply set_eq_sym in K10;apply ID_singleton_l with (C:=C) in K10;apply set_eq_sym in K9;apply ID_singleton_l with (C:=C) in K9.
  apply set_eq_sym in K9;apply set_eq_trans with (y:=ID (ι u) C);assumption.
Qed.

Lemma ID_subst_r : forall (A B C:N), A ≡ B -> (set_eq (ID C A)(ID C B)).
Proof.
intros A B C H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v)).
  -- split;apply N9.
  -- apply singular_eq_dec in H;assert (set_eq (ι u) (ι v)).
     --- apply set_eq_trans with (y:=A);[ assumption | apply set_eq_trans with (y:=B);[ assumption | apply set_eq_sym;assumption ]].
     --- rewrite H in H0;assumption.
- rewrite <-H in K10;apply set_eq_sym in K10;apply ID_singleton_r with (C:=C) in K10;apply set_eq_sym in K9;apply ID_singleton_r with (C:=C) in K9.
  apply set_eq_sym in K9;apply set_eq_trans with (y:=ID C (ι u));assumption.
Qed.

Lemma propag_EDT_in : forall A E E', η E (ED A E') -> exists F F', η F (el E) /\ η F (ET A) /\ η F' (el E') /\ η F' (ET A) /\ η F (ED A F').
Proof.
intros;exists E, E';hauto lq: on use: edt_is_ET, SinII unfold: η.
Qed.

Lemma propag_EDT_out : forall A E E', η E (ED A E') -> exists F F', η E (el F) /\ η F (ET A) /\ η E' (el F') /\ η F' (ET A) /\ η F (ED A F').
Proof.
intros;exists E, E';hauto lq: on use: edt_is_ET, SinII unfold: η.
Qed.


Lemma propag_ET_IT : forall A C E E' F F', η E (ED A E') /\ η E (IT C) /\ η E' (IT C) /\ η F (ED A F') -> exists X X', 
                                                                                               η X (el E) /\ η X (IT C) /\ η X' (el E') /\ η X' (IT C).
Proof.
intros;exists E, E';strivial use: SinII unfold: η.
Qed.

Lemma edt_to_it : forall A E E' F F', η E (ED A E') /\ η F (ED A F') /\ η F (pt E) /\ η F' (pt E') -> η F (IT E) /\ η F' (IT E').
Proof.
intros. destruct H as [H1 [H2 [H3 H4]]];split.
- rewrite intTanTo;rewrite diamExtTanTo in *;destruct H1 as [H1 [H5 H6]];destruct H2 as [H2 [H7 H8]];split.
  -- strivial use: extTanTo.
  -- split.
     --- strivial use: extTanTo.
     --- split.
         ---- assumption.
         ---- intros. destruct H as [J1 [J2 [J3 [J4 [J5 J6]]]]];clear H6 H8.
              rewrite extTanTo in H2;destruct H2 as [K1 [K2 [K3 K4]]];apply K4;split.
              ----- assumption.
              ----- split.
                    ------ assumption.
                    ------ split;[ assumption |
                           split;[ apply ET_is_ext in H1;apply ext_el with (P:=E);split;assumption |
                           split;[ assumption | apply ET_is_ext in H1;apply ext_el with (P:=E);split;assumption ]]].
- rewrite intTanTo;rewrite diamExtTanTo in *;destruct H1 as [H1 [H5 H6]];destruct H2 as [H2 [H7 H8]];split.
  -- strivial use: extTanTo.
  -- split.
     --- strivial use: extTanTo.
     --- split.
         ---- assumption.
         ---- intros. destruct H as [J1 [J2 [J3 [J4 [J5 J6]]]]];clear H6 H8.
              rewrite extTanTo in H7;destruct H7 as [K1 [K2 [K3 K4]]];apply K4;split.
              ----- assumption.
              ----- split.
                    ------ assumption.
                    ------ split;[ assumption | split;[
                           apply ET_is_ext in H5;apply ext_el with (P:=E');split;assumption |
                           split;[assumption | apply ET_is_ext in H5;apply ext_el with (P:=E');split;assumption ]]].
Qed.


Lemma edt_to_ext : forall A X Y, η X (ED A Y) -> η X (ext Y).
Proof.
intros A X Y H;rewrite diamExtTanTo in H;destruct H as [H1 [H2 H3]];apply H3;split.
- rewrite extTanTo in H1;destruct H1;assumption.
- split.
  -- rewrite extTanTo in H2;destruct H2;assumption.
  -- split;[ strivial use: SinII unfold: η | split;[ apply ET_is_ext;assumption | split;[ apply SinII;destruct H2;assumption | apply ET_is_ext;assumption ]]].
Qed.


(** A is concentric with B **)

Definition Concent (B:N) : N :=
    Caract (fun A:object => IF_then_else (In balls A /\ η B balls /\ ( (ι A) ≡ B \/
                                (η (ι A) (pt B) /\ exists P Q R S, η P (ED (ι A) Q) /\ η P (ID B Q) /\ η R (ED (ι A) S) /\ η R (ID B S) /\ ~(P ≡ R) /\ ~(P ≡ S)) \/
                                (η B (pt (ι A)) /\ exists P Q R S, η P (ED B Q) /\ η P (ID (ι A) Q) /\ η R (ED B S) /\ η R (ID (ι A) S) /\ ~(P ≡ R) /\ ~(P ≡ S)))
                                ) True False).


Lemma is_concent : forall A B, η A (Concent B) <-> (η A balls /\ η B balls /\ ( A ≡ B \/
                                 (η A (pt B) /\ exists P Q R S, η P (ED A Q) /\ η P (ID B Q) /\ η R (ED A S) /\ η R (ID B S) /\ ~(P ≡ R) /\ ~(P ≡ S)) \/
                                 (η B (pt A) /\ exists P Q R S, η P (ED B Q) /\ η P (ID A Q) /\ η R (ED B S) /\ η R (ID A S) /\ ~(P ≡ R) /\ ~(P ≡ S)))).
Proof.
intros A B;split.
- intro;solve_op_in_hyp H Concent x;destruct H2.
  -- destruct H1 as [[H2 [H3 H4]] H1];split.
     --- unfold η;split;[ unfold Individual;exists x;assumption | apply in_in_singleton with (A:=x);split;assumption ].
     --- split.
         ---- assumption.
         ---- destruct H4.
              ----- classical_left;unfold singular_eq in *;qauto depth: 4 l: on use: N22, rewl_singleton_in_η unfold: ι, set_eq.
              ----- destruct H4;[
                    classical_right;classical_left;destruct H4 as [H4 H7];split;[ strivial use: set_eq_sym, η_singleton unfold: η_inv |
                    clear H5 H6;destruct H7 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[
                    apply set_eq_sym in H;apply ED_singleton_r with (C:=E) in H;rewrite N22 in H;rewrite <-H;assumption | split;[ assumption |
                    split;[ apply set_eq_sym in H;apply ED_singleton_r with (C:=G) in H;rewrite N22 in H;rewrite <-H;assumption |
                    split;[ assumption | split;assumption ]]]]] |
                    classical_right;classical_right;destruct H4 as [H4 H7];split;[ apply rewl_pt_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ] |
                    clear H5 H6;destruct H7 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[ assumption |
                    split;[ apply set_eq_sym in H;apply ID_singleton_l with (C:=E) in H;rewrite N22 in H;rewrite <-H;assumption |
                    split;[ assumption |
                    split;[ apply set_eq_sym in H;apply ID_singleton_l with (C:=G) in H;rewrite N22 in H;rewrite <-H;assumption |
                    split;assumption ]]]]]].
  -- destruct H1;contradiction.
- intro;destruct H as [H1 [H2 H3]];unfold η;destruct H1 as [H1 H4];split.
  -- assumption.
  -- destruct H1 as [x H1];apply in_in_singleton with (A:=x);split.
     --- assumption.
     --- solve_op_in_goal Concent;split.
         ---- split.
              ----- apply incl_in_singleton with (Φ:=A);split;assumption.
              ----- split;[ assumption |
                    destruct H3;[ classical_left;hauto use: N10, N11, η_singleton_l, in_singleton_eq unfold: set_eq, incl, η, singular_eq, Individual |
                    destruct H0;[ classical_right;classical_left;destruct H0;split;[ strivial use: set_eq_sym, η_singleton unfold: η_inv |
                    clear H3 H5;destruct H6 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[
                    apply set_eq_sym in H1;apply ED_singleton_r with (C:=E) in H1;rewrite N22 in H1;rewrite H1;assumption |
                    split;[ assumption |
                    split;[ apply set_eq_sym in H1;apply ED_singleton_r with (C:=G) in H1;rewrite N22 in H1;rewrite H1;assumption |
                    split;[ assumption | split;assumption ]]]]] |
                    classical_right;classical_right;clear H4 H5;destruct H0 as [H4 H7];split;[
                    apply rewr_pt_singleton_in_η with (ϕ:=A);split;[ assumption | apply set_eq_sym;assumption ] |
                    destruct H7 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[ assumption |
                    split;[ apply set_eq_sym in H1;apply ID_singleton_l with (C:=E) in H1;rewrite N22 in H1;rewrite H1;assumption |
                    split;[ assumption | split;[
                    apply set_eq_sym in H1;apply ID_singleton_l with (C:=G) in H1;rewrite N22 in H1;rewrite H1;assumption |
                    split;assumption ]]]]]]]].
        ---- auto.
Qed.

Lemma concent_singleton : forall (B :N)(x :object), set_eq B (ι x) -> (set_eq (Concent (ι x))(Concent B)).
Proof. 
intros B x H0;unfold set_eq;intro y;apply propositional_extensionality;split. 
- intro H;unfold In in H;unfold Concent in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold Concent;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H10 [H3 H4]] H5];split.
              ----- split.
                    ------ assumption.
                    ------ split;[ apply rewl_singleton_in_η with (A:=x);split;assumption |
                           destruct H4;[
                           classical_left;hauto lq: on use: rewl_singleton_in_η, set_eq_equiv, eq_obj_singular_eq, set_eq_sym, singular_eq_singl, indiv_singleton unfold: η |
                           destruct H;[
                           classical_right;classical_left;destruct H;split;[ apply rewl_pt_singleton_in_η with (A:=x);split;assumption |
                           clear H2 H4 H6;destruct H7 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[ assumption |
                           split;[ apply ID_singleton_l with (C:=E) in H0;rewrite N22 in H0;rewrite <-H0;assumption |
                           split;[ assumption |
                           split;[ apply ID_singleton_l with (C:=G) in H0;rewrite N22 in H0;rewrite <-H0;assumption | split; assumption ]]]]] | 
                           classical_right;classical_right;destruct H;split;[ apply rewl_singleton_in_η with (A:=x);split;assumption |
                           clear H2 H4 H6;destruct H7 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[
                           apply ED_singleton_r with (C:=E) in H0;rewrite N22 in H0;rewrite <-H0;assumption |
                           split;[ assumption |
                           split;[ apply ED_singleton_r with (C:=G) in H0;rewrite N22 in H0;rewrite <-H0;assumption |
                           split;[ assumption | split;assumption ]]]]]]]].
              ----- auto.
      --- destruct H;contradiction.
- intro H;unfold In in H;unfold Concent in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3.
     --- unfold In;unfold Concent;unfold IF_then_else;simpl;apply propositional_extensionality;split.
         ---- intro;auto.
         ---- intro;classical_left;destruct H as [[H [H3 H4]] H5];split.
              ----- clear H2;split.
                    ------ assumption.
                    ------ split;[ apply rewr_singleton_in_η with (σ:=B);split;assumption | destruct H4;[
                           classical_left;hauto use: set_eq_equiv, N9, singular_eq_trans, set_eq_sym unfold: singular_eq, η  |
                           destruct H2;[ classical_right;classical_left;clear H4 H6;destruct H2;split;[
                           apply rewr_pt_singleton_in_η with (ϕ:=B);split;assumption |
                           destruct H4 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[ assumption |
                           split;[ apply ID_singleton_l with (C:=E) in H0;rewrite N22 in H0;rewrite H0;assumption |
                           split;[ assumption |
                           split;[ apply ID_singleton_l with (C:=G) in H0;rewrite N22 in H0;rewrite H0;assumption |
                           split;assumption ]]]]] |
                           classical_right;classical_right;destruct H;destruct H2;clear H4 H6;split;[
                           apply rewr_singleton_in_η with (σ:=B);split;assumption |
                           destruct H2 as [D [E [F [G [K1 [K2 [K3 [K4 [K5 K6]]]]]]]]];exists D, E, F, G;split;[
                           apply ED_singleton_r with (C:=E) in H0;rewrite N22 in H0;rewrite H0;assumption |
                           split;[ assumption |
                           split;[ apply ED_singleton_r with (C:=G) in H0;rewrite N22 in H0;rewrite H0;assumption |
                           split;[ assumption | split;assumption ]]]]]]]].
              ----- auto.
     --- destruct H;contradiction.
Qed.

Lemma concent_subst : forall A B, A ≡ B -> (set_eq (Concent A)(Concent B)).
Proof.
intros A B H1;assert (K9:=H1);apply singular_singleton in K9;assert (K11:=K9);destruct K9 as [[u K9][v K10]];apply singular_eq_dec in K11;
rewrite <-K11 in H1;assert (u = v).
- rewrite <-singular_eq_singl;assert (Individual (ι u) /\ Individual (ι v));[ split;apply N9 | apply singular_eq_dec in H;assert (set_eq (ι u) (ι v));sfirstorder ].
- qauto use: concent_singleton, set_eq_sym, set_eq_trans.
Qed.

Lemma Concent_refl : forall A, η A balls -> η A (Concent A).
Proof.
hauto use: is_concent unfold: incl, singular_eq, η.
Qed.


Lemma Concent_sym : forall A B, η A (Concent B) -> η B (Concent A).
Proof.
intros A B H;rewrite is_concent in *;hfcrush use: set_eq_sym, singular_eq_eq_obj, singular_eq_dec unfold: ED, η, ID.
Qed.

Lemma ext_not_eq : forall x y, η x (ext y) -> ~(x ≡ y).
Proof.
hauto lq: on use: singular_eq_eq_obj, SinCXXXII.
Qed.

Lemma concent_to_el : forall A B, η A (Concent B) -> η A (el B) \/ η B (el A).
Proof.
hcrush use: SinXVIII, LejT0, singular_eq_eq_obj, singular_eq_sym, is_concent unfold: η.
Qed.

Lemma concent_to_pt : forall A B, η A (Concent B) -> A ≡ B \/ η A (pt B) \/ η B (pt A).
Proof.
hauto lq: on use: is_concent.
Qed.

Axiom concent_trans : forall A B C, η A (Concent B) /\ η B (Concent C) -> η A (Concent C).

Lemma concent_not_ext : forall A B, η A (Concent B) -> ~(η B (ext A)).
Proof.
hauto lq: on use: SinXXXII, SinCXV, concent_to_el.
Qed.

Lemma concent_sets : forall A B, η A (Concent B) -> (Concent A) ≈ (Concent B).
Proof.
intros A B H;assert (H':=H);apply Concent_sym in H';unfold weak_eq;intro C;split;[
intro H2;apply concent_trans with (B:=A);split;assumption | intro H2;apply concent_trans with (B:=B);split;assumption ].
Qed.

(**  Saturated subsphere **)

Definition sat_subsphere (B:N) : N :=  Caract (fun A:object => IF_then_else ( η (ι A) balls /\ η B Solid /\ η (ι A) (el B) /\
                                               (forall x, η x balls /\ η (ι A)(el x) -> η (ι A) x)) True False).

Lemma is_sat_subsphere : forall A B, η A (sat_subsphere B) <-> (η A balls /\ η B Solid /\ η A (el B) /\ (forall x, η x balls /\ η A (el x) -> η A x)).
Proof.
intros A B;split.
- intro;assert (H0:=H);solve_op_in_hyp H sat_subsphere x;destruct H3.
  -- destruct H2 as [[H2 [H3 [H4 H5]]] H6];split;[
     apply rewl_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ] |
     split;[ assumption |
     split;[ apply rewl_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ] |
     intros y [H7 H8];apply rewl_singleton_in_η with (A:=x);split;[
     apply H5;split;[ assumption | apply rewr_singleton_in_η with (σ:=A);split;[ assumption | apply set_eq_sym;assumption ]] |
     apply set_eq_sym;assumption ]]]].
  -- destruct H2;contradiction. 
- intro;destruct H as [H3 [H4 [H5 H6]]];assert (H10:=H3);unfold η;split.
  -- destruct H3;assumption.
  -- destruct H3 as [[x H1] H2];apply in_in_singleton with (A:=x);split;[ assumption |
     solve_op_in_goal sat_subsphere;split;[
     apply set_eq_sym in H1;split;[ apply rewr_singleton_in_η with(σ:=A);split;assumption |
     split;[ assumption | split;[ qauto use: N9 unfold: η, incl, set_eq |
     intros y [H7 H8];apply rewr_singleton_in_η with (σ:=A);split;[
     apply H6;split;[ assumption | apply rewl_singleton_in_η with (A:=x);split;assumption ] | assumption ]]]] | auto ]].
Qed.

(** Specification of a point as a saturated subsphere of Gspace relative to ball A **)

Definition Point : N :=
       Caract (fun A:object => IF_then_else (η (ι A) (sat_subsphere Gspace) /\ η (ι A) (klass (Concent (ι A)))) True False).


Lemma Point_def : forall A, η A Point <-> η A (sat_subsphere Gspace) /\ η A (klass (Concent A)).
Proof.
Proof.
intro A;split.
- intro;unfold η in H;destruct H as [[x H] H20];assert (H21:set_eq (ι x) A /\ incl A Point).
  -- split;assumption.
  -- apply incl_in_singleton in H21;unfold In in H21;unfold Point in H21;unfold IF_then_else in H21;simpl in H21;assert (H1:True);[
     auto |
     rewrite <-H21 in H1;clear H21;destruct H1;[ destruct H0 as [[H0 H1] H3];split;[
     apply rewl_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ] |
     apply set_eq_sym in H;assert (H2:=H);apply concent_singleton in H;apply MieT16 with (B:=Concent (ι x));split;[
     apply rewl_singleton_in_η with (A:=x);split;assumption | assumption ]] |
     destruct H0;contradiction ]].
- intros [H H'];assert (H0:=H);destruct H as [[x H] H1];unfold η;split.
  -- destruct H0;assumption.
  -- apply in_in_singleton with (A:=x);split;[ assumption |
     unfold In;unfold Point;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto |
     intro;classical_left;split;[
     split;[ hauto lq: on use: N9, set_eq_equiv, OntoT7 unfold: η |
     apply set_eq_sym in H;assert (H4:=H);apply concent_singleton in H; apply MieT16 with (B:=Concent A);split;[
     qauto use: N9 unfold: η, incl, set_eq | scongruence unfold: set_eq ]] | auto ]]].
Qed.

Lemma equiv_points : forall A B, η A Point /\ η B Point /\ η A (Concent B) -> η A B.
Proof.
intros A B [H1 [H2 H3]];rewrite Point_def in *;rewrite is_sat_subsphere in *;destruct H1 as [H1 K0];destruct H1 as [K1 [K2 [K3 K4]]].
destruct H2 as [H2 K5];destruct H2 as [K6 [K7 [K8 K9]]];rewrite Klass in K0;rewrite Klass in K5;hfcrush.
Qed.


Lemma sat_points : forall A, η A Point -> exists B, η B (sat_subsphere Gspace) /\ η A (Concent B).
Proof.
hfcrush use: Concent_refl, Point_def, is_sat_subsphere. 
Qed.

Lemma any_ball_el_of_its_concent_klass : forall A, η A balls -> η A (el (klass (Concent A))).
Proof.
intros A H;apply MieT22;apply Concent_refl;assumption.
Qed.

Lemma sat_balls : forall X, η X balls -> (exists P, η P (sat_subsphere Gspace) /\ η P (klass (Concent X)) -> η P Point).
Proof.
strivial use: Point_def.
Qed.

Lemma klass_concent : forall Y, η Y balls -> exists X, (η X (sat_subsphere Gspace) /\ η Y (pt X) /\ η Y (Concent X) -> η X (klass (Concent X))).
Proof.
intros;assert (H1:=H);apply LejT1 in H;destruct H as [X H];exists X;intros [H2 [H3 H4]];rewrite is_sat_subsphere in H2.
destruct H2 as [K1 [K2 [K3 K4]]];rewrite part in H3;destruct H3;assert (set_eq X Y);[ apply SinLI;split;assumption | contradiction ].
Qed.

Lemma points_el_of_Gspace : forall A, η A Point -> η A (el Gspace).
Proof.
hfcrush use: is_sat_subsphere, Point_def.
Qed.

Lemma pt_is_ball : forall A, η A Point -> η A balls.
Proof.
hcrush use: is_concent, sat_points.
Qed.

(** Surface of a Ball **)

Definition on_surface (A :N) : N :=
 Caract (fun P:object => IF_then_else (η A balls /\ η (ι P) Point /\ (forall X, η X balls /\ η X (Concent (ι P)) -> ~η X (el A) /\ ~η X (ext A))) True False).

Lemma in_surface : forall A B, η B (on_surface A) <-> (η A balls /\ η B Point /\ (forall X, η X balls /\ η X (Concent B) -> ~η X (el A) /\ ~η X (ext A))).
Proof.
intros A B;split.
- intro;solve_op_in_hyp H on_surface x;destruct H2.
  -- destruct H1 as [[H3 [H4 H5]] H2];split;[ assumption |
     split;[ strivial use: set_eq_sym, η_singleton unfold: η_inv |
     intros C [H6 H7];apply H5;split;[ assumption |
     apply set_eq_sym in H;apply concent_singleton in H;rewrite N22 in H;rewrite H;assumption ]]].
  -- destruct H1;contradiction.
- intro;destruct H as [H3 [H4 H5]];unfold η;split.
  -- destruct H4;assumption.
  -- assert (H10:=H4);destruct H4 as [[x H1] H2];apply in_in_singleton with (A:=x);split;[ assumption |
     solve_op_in_goal on_surface;split;[
     split;[ assumption |
     split;[ apply rewr_singleton_in_η with (σ:=B);split;[ assumption | apply set_eq_sym;assumption ] |
     intros C [H6 H7];apply H5;split;[ assumption |
     apply set_eq_sym in H1;apply concent_singleton in H1;hauto use: set_eq_sym, N22 ]]] | auto ]].
Qed.

Lemma surface_singleton : forall (B :N)(x :object), set_eq B (ι x) -> (set_eq (on_surface (ι x))(on_surface B)).
Proof. 
intros B x H0;unfold set_eq;intro y;apply propositional_extensionality;split.
- intro H;unfold In in H;unfold on_surface in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3;[
     unfold In;unfold on_surface;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto |
     intro;classical_left;destruct H as [[H10 [H3 H4]] H5];split;[
     split;[ apply rewl_singleton_in_η with (A:=x);sauto | split;[ assumption |
     intros X H6;apply H4 in H6;destruct H6;split;[ intro;apply H;apply rewr_el_singleton_in_η with (ϕ:=B);sauto |
     intro;apply H6;apply rewr_ext_singleton_in_η with (ϕ:=B);sauto ]]] | auto ]] | destruct H;contradiction ].
- intro H;unfold In in H;unfold on_surface in H;unfold IF_then_else in H;simpl in H;assert (H3:True).
  -- auto.
  -- rewrite <-H in H3;clear H;destruct H3;[
     unfold In;unfold on_surface;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto |
     intro;classical_left;destruct H as [[H10 [H3 H4]] H5];split;[
     split;[ apply rewr_singleton_in_η with (σ:=B);sauto | split;[ assumption |
     intros X H6;apply H4 in H6;destruct H6;split;[ intro;apply H;apply rewl_el_singleton_in_η  with (A:=x);sauto |
     intro;apply H6;apply rewl_ext_singleton_in_η  with (A:=x);sauto ]]] | auto ]] | destruct H;contradiction ].
Qed.

Lemma ball_not_on_its_surface : forall A, ~η A (on_surface A).
Proof.
intros A H;rewrite in_surface in H;destruct H as [H1 [H2 H3]];specialize (H3 A);assert (η A balls ∧ η A (Concent A));[
split;[ assumption | apply Concent_refl;auto ] | hauto lq: on use: Point_def, propositional_extensionality, Klass unfold: In, incl, η ].
Qed.


(** Points B C are equidistant from ball A  **)

Definition Equid (B C:N) : N :=
    Caract (fun A:object => IF_then_else (η B Point /\ η C Point /\ η B (on_surface (ι A)) /\ η C (on_surface (ι A))) True False).

Lemma equidistant : forall A B C, η A (Equid B C) <-> (η B Point /\ η C Point /\ η B (on_surface A) /\ η C (on_surface A)).
Proof.
intros A B C;split.
- intro;unfold η in H;destruct H as [[x H] H1];assert (H2:In (Equid B C) x).
  -- apply incl_in_singleton with (Φ:=A);split;assumption.
  -- unfold In in H2;unfold Equid in H2;unfold IF_then_else in H2;simpl in H2;assert (H3:True);[ auto |
     rewrite <-H2 in H3;destruct H2;destruct H3;[
     destruct H0 as [[H2 [H3 [H4 H5]]] H0];split;[ assumption | split;[ assumption |
     split;repeat ( apply set_eq_sym in H;apply surface_singleton in H;rewrite N22 in H;rewrite <-H;assumption )]] |
     destruct H0;contradiction ]].
- intros [H1 [H2 [H3 H4]]];assert (H0:=H3);unfold η;rewrite in_surface in H0;destruct H0;split.
  -- destruct H;assumption.
  -- clear H0;destruct H;destruct H as [x H];apply in_in_singleton with (A:=x);split;[ assumption |
     unfold In;unfold Equid;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto |
     intro;classical_left;split;[ split;[ assumption |
     split;[ assumption | split;repeat ( hcrush use: surface_singleton, set_eq_sym, N22 )]] | auto ]]].
Qed.


Lemma Equid_refl : forall A B, η A Point /\ η B (on_surface A) -> η A (Equid B B).
Proof.
intros A B [H1 H2];rewrite equidistant;assert (H10:=H2);rewrite in_surface in H2;destruct H2 as [H2 [H3 H4]];split;[
assumption | split;[ assumption | split;assumption ]].
Qed.

Lemma Equid_sym : forall A B C, η A (Equid B C) -> η A (Equid C B).
Proof.
hcrush use: equidistant.
Qed. 

Lemma Equid_trans : forall A B C D, η A Point -> exists A', η A' balls /\ η A' (Equid B C) /\ η A' (Equid C D) -> η A' (Equid B D).
Proof.
hauto lq: on use: equidistant, ball_not_on_its_surface.
Qed.

(** Betweeness between balls **)

Definition Btw (B C:N) : N :=
    Caract (fun A:object => IF_then_else (In balls A /\ exists D E F, η D (Concent B) /\ η E (Concent C) /\ η F (Concent (ι A)) /\ η D (ED F E)) True False).

Lemma betw_balls : forall A B C, η A (Btw B C) <-> (η A balls /\ exists D E F, η D (Concent B) /\ η E (Concent C) /\ η F (Concent A) /\ η D (ED F E)).
Proof.
intros A B C;split.
- intro;destruct H;destruct H as [x H];assert (H2:In (Btw B C) x).
  -- apply incl_in_singleton with (Φ:=A);split;assumption.
  -- unfold In in H2;unfold Btw in H2;unfold IF_then_else in H2;simpl in H2;assert (H3:True);[ auto |
     rewrite <-H2 in H3;destruct H2;destruct H3;[ destruct H1 as [[H2 H3] H1];split;[
     apply rewl_singleton_in_η with (A:=x);split;[ rewrite <-N12;assumption | apply set_eq_sym;sauto ] |
     destruct H3 as [D [E [F [K1 [K2 [K3 K4]]]]]];exists D, E, F;split;[
     assumption | split;[ assumption | split;[
     apply set_eq_sym in H;apply concent_singleton in H;rewrite N22 in H;rewrite <-H;assumption |
     assumption ]]]] | destruct H1;contradiction ]].
- intros [H1 H2];unfold η;split.
  -- destruct H1;assumption.
  -- destruct H1;destruct H as [x H];apply in_in_singleton with (A:=x);split;[ assumption |
     unfold In;unfold Btw;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto |
     intro;classical_left;split;[ split;[ apply incl_in_singleton with (Φ:=A);sauto |
     destruct H2 as [D [E [F [K1 [K2 [K3 K4]]]]]];exists D, E, F;split;[ assumption | split;[ assumption |
     split;[ apply set_eq_sym in H;apply concent_singleton in H;rewrite N22 in H;rewrite H;assumption |
     assumption ]]]] | auto ]]].
Qed.


Lemma btw_is_ED : forall A B C, η B (ED A C) -> η A (Btw B C).
Proof.
intros A B C H;rewrite betw_balls;split.
- rewrite diamExtTanTo in H;destruct H;rewrite extTanTo in H;destruct H as [H1 [H2 H3]];assumption.
- exists B, C, A;assert (H0:=H);rewrite diamExtTanTo in H;destruct H as [H1 [H2 H3]];rewrite extTanTo in H1, H2;strivial use: Concent_refl.
Qed.

(** Betweeness between points **)

Definition btw (A B:N) : N :=
    Caract (fun C:object => IF_then_else (η A Point /\ η B Point /\ exists A' B', η A' balls /\ η B' balls /\ η A' (Concent A) /\ 
                                                                                     η B' (Concent B) /\ η (ι C) (Btw A' B')) True False).

Lemma betw_pts : forall A B C, η C (btw A B) <-> (η A Point /\ η B Point /\ exists A' B', η A' balls /\ η B' balls /\ η A' (Concent A) /\ 
                                                                                     η B' (Concent B) /\ η C (Btw A' B')).
Proof.
intros A B C;split.
- intro;destruct H;destruct H as [x H];assert (H2:In (btw A B) x).
  -- apply incl_in_singleton with (Φ:=C);split;assumption.
  -- unfold In in H2;unfold btw in H2;unfold IF_then_else in H2;simpl in H2;assert (H3:True);[ auto |
     rewrite <-H2 in H3;clear H2;destruct H3;[
     destruct H1 as [[H2 [H3 H4]] H1];split;[ assumption |
     split;[ assumption | destruct H4 as [D [E [K1 [K2 [K3 [K4 K5]]]]]];exists D, E;split;[ assumption |
     split;[ assumption | split;[ assumption | split;[ assumption |
     apply rewl_singleton_in_η with (A:=x);split;[ assumption | apply set_eq_sym;assumption ]]]]]]] |
     destruct H1;contradiction ]].
- intro;destruct H as [H1 [H2 H3]];unfold η;destruct H3 as [D [E [K1 [K2 [K3 [K4 K5]]]]]];split.
  -- rewrite betw_balls in K5;destruct K5;destruct H;assumption.
  -- assert (K6:=K5);rewrite betw_balls in K5;destruct K5;destruct H;destruct H as [x H];apply in_in_singleton with (A:=x);split;[
     assumption | unfold In;unfold btw;unfold IF_then_else;simpl;apply propositional_extensionality;split;[ intro;auto |
     intro;classical_left;clear H5 H0;split;[ split;[ assumption | split;[ assumption |
     exists D, E;split;[ assumption | split;[ assumption | split;[ assumption | split;[ assumption |
     apply rewr_singleton_in_η with (σ:=C);split;[ assumption | apply set_eq_sym;assumption ]]]]]]] | auto ]]].
Qed.


Lemma T4 : forall A B D, η D (Btw A B) -> ~η A (Concent B).
Proof.
intros A B D. rewrite Contra;intro;rewrite <-notnot in H;intro H1;rewrite betw_balls in H1;destruct H1 as [H1 [P [Q [R [H2 [H3 [H4 H5]]]]]]]. 
assert (η P (Concent B)).
- apply concent_trans with (B:=A);sauto.
- assert (η P (Concent Q));[ apply concent_trans with (B:=B);hfcrush use: Concent_sym |
  apply edt_to_ext in H5;apply concent_not_ext in H6;apply SinXXXII in H5;contradiction ].
Qed.

Lemma T5 : forall A B D P Q R, η R (Concent D) /\ η P (Concent A) /\ η Q (Concent B) -> (η D (Btw A B) <-> η R (Btw P Q)).
Proof.
intros A B C P Q R [H1 [H2 H3]];split.
- intro H4;rewrite betw_balls in *;split;[ rewrite is_concent in H1;destruct H1 as [H1 [H0 K0]];assumption |
  apply Concent_sym in H1;apply Concent_sym in H2;apply Concent_sym in H3;destruct H4 as [H4 [D [E [F [K1 [K2 [K3 K4]]]]]]];exists D, E, F;split;[
  apply concent_trans with (B:=A);split;assumption |
  split;[ apply concent_trans with (B:=B);split;assumption | split;[ apply concent_trans with (B:=C);split;assumption | assumption ]]]].
- intro H4;rewrite betw_balls in *;split;[ rewrite is_concent in H1;destruct H1 as [H1 [H0 K0]];assumption |
  destruct H4 as [H4 [D [E [F [K1 [K2 [K3 K4]]]]]]];exists D, E, F;split;[
  apply concent_trans with (B:=P);sauto | split;[ apply concent_trans with (B:=Q);sauto |
  split;[ apply concent_trans with (B:=R);sauto | assumption ]]]].
Qed.


Lemma T6 : forall A B D, η D (btw A B) <-> (η D balls /\ (Concent D) ⊆ (btw A B)).
Proof.
intros A B D;split.
- intro;split.
  -- rewrite betw_pts in H;destruct H as [H2 [H3 H4]];destruct H4 as [E [F [K1 [K2 [K3 [K4 K5]]]]]];rewrite betw_balls in K5;destruct K5;assumption.
  -- assert (H10:=H);unfold weakInclusion;intros C H1;rewrite betw_pts in H;destruct H as [H2 [H3 H4]];rewrite betw_pts;split;[ assumption |
     split;[ assumption | destruct H4 as [E [F [K1 [K2 [K3 [K4 K5]]]]]];exists E, F;split;[ assumption |
     split;[ assumption | split;[ assumption | split;[ assumption |
     assert (η C (Concent D) /\ η E (Concent A) /\ η F (Concent B));[ hfcrush |
     apply T5 in H;rewrite <-H;rewrite betw_balls;rewrite betw_balls in K5;destruct K5;destruct H4 as [P [Q [R [K6 [K7 [K8 K9]]]]]];split;[
     assumption | exists P, Q, R;split;[ apply concent_trans with (B:=E);split;assumption |
     split;[apply concent_trans with (B:=F);split;assumption | sauto ]]]]]]]]]].
- hfcrush use: Concent_refl unfold: weakInclusion.
Qed.

Lemma T7 : forall A B, (btw A B) ≈ (btw B A).
Proof.
intros A B;unfold weak_eq;intro C;split;(
intro;rewrite betw_pts in *;destruct H as [H1 [H2 H3]];destruct H3 as [D [E [K1 [K2 [K3 [K4 K5]]]]]];split;[
assumption | split;[ assumption | exists E, D;split;[ assumption | split;[ assumption | split;[ assumption | split;[ assumption |
rewrite betw_balls in *;destruct K5 as [K5 [F [G [H [K6 [K7 [K8 K9]]]]]]];split;[
assumption | exists G, F, H;hfcrush use: ED_sym ]]]]]]] ).
Qed.


(** Interior point **)

Definition InteriorPoint (B :N) : N :=
    Caract (fun A:object => IF_then_else (η B Solid /\ η (ι A) Point /\ exists C, η C (Concent (ι A)) /\ η C (el B)) True False).

Lemma is_int_point : forall A B, η A (InteriorPoint B) <-> (η B Solid /\ η A Point /\ exists C, η C (Concent A) /\ η C (el B)).
Proof.
intros A B;split.
- intro;solve_op_in_hyp H InteriorPoint x;destruct H2;[
  destruct H1 as [[H3 [H4 H5]] H2];split;[ assumption | split;[
  apply rewl_singleton_in_η  with (A:=x);split;[ assumption | apply set_eq_sym;assumption ] |
  destruct H5 as [C [H5 H6]];exists C;split;[
  apply set_eq_sym in H;apply concent_singleton in H;rewrite N22 in H;rewrite <-H;assumption |
  assumption ]]] | destruct H1;contradiction ].
- intro;destruct H as [H1 [H2 [C [H3 H4]]]];unfold η;split;[ destruct H2;assumption |
  assert (H10:=H2);destruct H2 as [[x H2] H5];apply in_in_singleton with (A:=x);split;[ assumption |
  solve_op_in_goal InteriorPoint;split;[ split;[ assumption |
  split;[ apply rewr_singleton_in_η with (σ:=A);split;[ assumption | apply set_eq_sym;assumption ] |
  exists C;split;[ apply set_eq_sym in H2;apply concent_singleton in H2;rewrite N22 in H2;rewrite H2;assumption | 
  assumption ]]] | auto ]]].
Qed.

Axiom Tarski_3 : forall A, η A (klass (InteriorPoint A)) <-> η A Solid.
Axiom Tarski_4: forall A B, η A Solid /\ η B (el A) -> η B Solid.

Lemma MinimalPoint : forall A,  η A Gspace -> exists B, η B (InteriorPoint A).
Proof.
hfcrush use: MieT25, whole_space_is_a_solid, Tarski_3 unfold: exists_at_least.
Qed.

Lemma interior_propag : forall A C,  η C Solid /\ η A (el C) -> forall B, η B (InteriorPoint A) -> η B (InteriorPoint C).
Proof.
qauto depth: 4 l: on use: SinXXIII, is_int_point.
Qed.

Lemma solid_to_incl : forall A B,  η A Solid /\ η B (el A) -> (InteriorPoint B) ⊆ (InteriorPoint A).
Proof.
hauto use: propositional_extensionality, interior_propag unfold: In, incl, el, weakInclusion, InteriorPoint, η.
Qed.

Lemma el_to_incl :  forall A B, η A Solid /\ η B Solid -> η A (el B) -> (InteriorPoint A) ⊆ (InteriorPoint B).
Proof.
strivial use: solid_to_incl.
Qed.


Lemma Tarski_4b : forall A B, η A Solid /\ η B Solid -> (InteriorPoint A) ⊆ (InteriorPoint B) -> η A (el B).
Proof.
intros A B [H1 H2] H3;assert ((klass (InteriorPoint A)) ⊆ (el (klass (InteriorPoint B))));[ apply L18;assumption |
assert (K1:=H1);assert (K2:=H2);rewrite <-Tarski_3 in H1;rewrite <-Tarski_3 in H2;assert (H10:=H1);
apply MieT24 in H1;apply MieT24 in H2;assert (Individual B /\ Individual (klass (InteriorPoint B)));[
strivial unfold: η, InteriorPoint, klass, singular_eq | apply singular_eq_dec in H0;rewrite <-H0 in H2;clear H0;
assert (η A (el (klass (InteriorPoint B))));[ fcrush | apply MieT18 with (B:=klass (InteriorPoint B));split;[ | 
apply set_eq_sym];assumption ]]].
Qed.

Lemma equiv_el_incl : forall A B, η A Solid /\ η B Solid -> η A (el B) <-> (InteriorPoint A) ⊆ (InteriorPoint B).
Proof.
strivial use: solid_to_incl, Tarski_4b.
Qed.

Lemma Tarski_4c : forall A B, η A balls /\ η B (el A) -> exists C, η C balls /\ η C (el B).
Proof.
intros A B [H1 H2];assert (H3:forall A B, η A Solid /\ η B (el A) -> η B Solid);[ apply Tarski_4 |
specialize (H3 A B);repeat rewrite solid_as_coll in *;assert (H0:=H1);apply SinXIII in H1;assert (η B (coll balls));[
apply H3;split;assumption | apply SinXXI in H;sfirstorder ]].
Qed.



End Mereo_signature.
